/*
** Apple Macintosh Developer Technical Support
**
** Program:         texteditcontrol.c
** Written by:      Eric Soldan
** Based on:        TESample, by Bryan Stearns
** Suggestions by:  Forrest Tanaka, Dave Radcliffe
**
** Copyright  1990-1991 Apple Computer, Inc.
** All rights reserved.
*/

/* This is a control implementation of TextEdit.  The advantages to this are:
**
** 1) Makes using TextEdit in a non-dialog window easy.
** 2) The TextEdit record is automatically associated with the window, since
**    it is in the window's control list.
** 3) The TextEdit control can have scrollbars associated with it, and these
**    are also kept in the window's control list.
** 4) Updating of the TextEdit record is much simpler, since all that is
**    necessary is to draw the control (or all the window's controls with
**    a DrawControls call).
** 5) There are simple calls to handle TextEdit events.
** 6) Undo is already supported.
** 7) A document length can be specified.  This length will not be exceeded
**    when editing the TextEdit record.
** 8) When you close the window, the TextEdit record is disposed of.
**    (This automatic disposal can easily be defeated.)
**
**
** To create a TextEdit control, you only need a single call.  For example:
**
**	CTENew(rViewCtl,			Resource ID of view control for TextEdit control.
**		   window,				Window to hold TERecord.
**		   &teHndl,				Return handle for TERecord.
**		   &ctlRect,			Rect for TextEdit view control.
**		   &destRect,			destRect for TERecord
**		   &viewRect,			viewRect for TERecord
**		   &borderRect,			Used to frame a border.
**		   32000,				Max size for TERecord text.
**		   cteVScrollLessGrow	TERecord read-write, with vertical scroll
**								that leaves space for grow box.
**	);
**
** If you create a TextEdit control that is read-only, you will not be able
** to edit it, of course.  There will also be no blinking caret for that
** TextEdit control.  You will be able to select text and copy to the
** clipboard, but that is all.

** Simply create destRect, viewRect, and borderRect appropriately, and
** then call CTENew (which stands for Control TENew).  If teHndl is returned
** nil, then CTENew failed.  Otherwise, you now have a TextEdit control in
** the window.
**
** NOTE: There is a TextEdit bug (no way!!) such that you may need to set the
**       viewRect right edge 2 bigger than the right edge of destRect.  If you
**       do not do this, then there will be some clipping on the right edge in
**       some cases.  Of course, you may want this.  You may want horizontal
**       scrolling, and therefore you would want the destRect substantially
**       larger than the viewRect.  If you don't want horizontal scrolling,
**       then you probably don't want any clipping horizontally, and therefore
**       you will need to set destRect.right 2 less than viewRect.right.
**
**
** If the CTENew call succeeds, you then have a TextEdit control in your
** window.  It will be automatically disposed of when you close the window.
** If you don't waht this to happen, then you can detatch it from the
** view control which owns it.  To do this, you would to the following:
**
**  viewCtl = CTEViewFromTE(theTextEditHndl);
**  if (viewCtl) SetCRefCon(viewCtl, nil);
**
** The view control keeps a reference to the TextEdit record in the refCon.
** If the refCon is cleared, then the view control does nothing.  So, all that
** is needed to detatch a TextEdit record from a view control is to set the
** view control's refCon nil.  Now if you close the window, you will still
** have the TextEdit record.
**
**
** To remove a TextEdit control completely from a window, you make one call:
**
**  CTEDispose(theTextEditHndl);
**
** This disposes of the TextEdit record, the view control, and any scrollbar
** controls that were created when the TextEdit control was created with
** the call CTENew.
**
**
** Events for TextEdit record are handled nearly automatically.  You can
** make one of 3 calls:
**
**  CTEClick(eventPtr);
**  CTEEvent(eventPtr);
**  CTEKey(eventPtr);
**
** In each case, if the event was handled, true is returned.  CTEEvent simply
** calls either CTEClick or CTEKey, whichever is appropriate.
**
**
** Another call you will want to use is CTEEditMenu.  This is used to set the
** state of cut/copy/paste/clear for TextEdit controls.  It checks the active
** control to see if text is selected, if the control is read-only, etc.
** Based on this information, it sets cut/copy/paste/clear either active
** or inactive.  If any menu items are set active, it returns true.
**
**
** One more high-level call is CTEUndo().  In response to an undo menu item
** being selected by the user, just call CTEUndo(), and the edits the user
** has made will be undone.  (This includes undoing an undo.)
**
**
** The last high-level call is CTEClipboard.  Call it when you want to do a
** cut/copy/paste/clear for the active TextEdit control.  The value to pass
** is as follows:
**
**  2: cut
**  3: copy
**  4: paste
**  5: clear
**
** These are the same values you would pass to a DA for these actions.
*/



/*****************************************************************************/



#ifndef __TEXTEDITCONTROL__
#include "TextEditControl.h"
#endif

#ifndef __CONTROLS__
#include <Controls.h>
#endif

#ifndef __ERRORS__
#include <Errors.h>
#endif

#ifndef __EVENTS__
#include <Events.h>
#endif

#ifndef __MEMORY__
#include <Memory.h>
#endif

#ifndef __MENUS__
#include <Menus.h>
#endif

#ifndef __OSUTILS__
#include <OSUtils.h>
#endif

#ifndef __RESOURCES__
#include <Resources.h>
#endif

#ifndef __SCRAP__
#include <Scrap.h>
#endif



/*****************************************************************************/



#define kLeftArrow	28
#define kRightArrow	29
#define kUpArrow	30
#define kDownArrow	31


typedef struct cdefRsrcJMP {
	long	moveInst;
	long	jsrInst;
	short	jmpInst;
	long	jmpAddress;
} cdefRsrcJMP;
typedef cdefRsrcJMP *cdefRsrcJMPPtr, **cdefRsrcJMPHndl;



/*****************************************************************************/



static pascal long	CTECtl(short varCode, ControlHandle ctl, short msg, long parm);
static short		theViewID;
static Boolean		canGoSlow;
static Handle		scrollProc = nil;



/*****************************************************************************/



static TEHandle			gActiveTEHndl;
	/* Currently active TextEdit record.  (nil if none active.) */

static TEHandle			gFoundTEHndl;
	/* Global value used to return info from the TextEdit control proc. */

static ControlHandle	gFoundViewCtl;
	/* Global value used to return info from the TextEdit control proc. */

static pascal void		VActionProc(ControlHandle scrollCtl, short part);
static pascal void		HActionProc(ControlHandle control, short part);
static void				AdjustOneScrollValue(TEHandle teHndl, ControlHandle ctl, Boolean vert);

ClikLoopProcPtr		gDefaultClikLoop;
	/* The clikLoop TextEdit wants to use.  Our custom clikLoop must call
	** this as well.  The default TextEdit clikLoop is stored here.
	*/

#define kTELastForWind	1
#define kCrChar			13


/*****************************************************************************/
/*****************************************************************************/



/* Activate this TextEdit record.  If another is currently active, deactivate
** that one.  The view control for this TextEdit record is also flagged to
** indicate which was the last active one for this window.  If the previous
** active TextEdit record was in the same window, then flag the old one off
** for this window.  The whole point for this per-window flagging is so that
** activate events can reactivate the correct TextEdit control per window.
*/

#pragma segment Controls
void	CTEActivate(Boolean active, TEHandle teHndl)
{
	WindowPtr		tePort, oldPort;
	ControlHandle	viewCtl;
	TEHandle		te;
	CTEDataHndl		teData;

	if (!teHndl) return;

	if (!active) {
		GetPort(&oldPort);
		SetPort(tePort = (*teHndl)->inPort);
		TEDeactivate(teHndl);
		if (teHndl == gActiveTEHndl) gActiveTEHndl = nil;
		viewCtl = CTEViewFromTE(teHndl);
		CTEUpdate(teHndl, viewCtl, true);
		SetPort(oldPort);
		return;
	}

	if (!(viewCtl = CTEViewFromTE(teHndl))) return;

	teData = (CTEDataHndl)(*viewCtl)->contrlData;
	(*teData)->mode |= cteActive;
	GetPort(&oldPort);
	SetPort(tePort = (*teHndl)->inPort);
	TEActivate(gActiveTEHndl = teHndl);
	CTEUpdate(teHndl, viewCtl, true);
	SetPort(oldPort);
		/* Let TextEdit know that it is supposed to be active. */

	for (viewCtl = nil;;) {
		viewCtl = CTENext(tePort, &te, viewCtl);
		if (!viewCtl) break;
		if (te != teHndl) {
			teData = (CTEDataHndl)(*viewCtl)->contrlData;
			(*teData)->mode &= (0xFFFF - cteActive);
			CTEActivate(false, te);
		}
	}
}



/*****************************************************************************/



/* Select or reselect a range of text without flashing or lurching. */

#pragma segment Controls
void	CTEFakeClick(short newStart, short newEnd, Boolean extend, TEHandle teHndl)
{
	WindowPtr	oldPort;
	short		lineHeight, oldStart, oldEnd, newLoc, dx, dy, selTop, selBot;
	Point		fakeClick;
	Rect		viewRct;
#ifdef THINK_PRE_5
	long		tempLong;
#endif

	if (!teHndl) return;

	if (newStart < 0) newStart = 0;
	if (newEnd > (*teHndl)->teLength) newEnd = (*teHndl)->teLength;

	oldStart = (*teHndl)->selStart;
	oldEnd   = (*teHndl)->selEnd;
	if ((extend) && (oldStart == newStart) && (oldEnd == newEnd)) return;

	GetPort(&oldPort);
	SetPort((*teHndl)->inPort);
	TEAutoView(false, teHndl);
	lineHeight = (*teHndl)->lineHeight;

	newLoc = (newStart < oldStart) ? newStart : newEnd;

#ifndef THINK_PRE_5
		fakeClick = TEGetPoint(newLoc, teHndl);
#else
		tempLong = TEGetPoint(newLoc, teHndl);
		BlockMove(&tempLong, &fakeClick, 4L);
#endif
	fakeClick.v -= (lineHeight >> 1);
	if (newLoc != TEGetOffset(fakeClick, teHndl)) fakeClick.v += lineHeight;

	(*teHndl)->clickTime = 0L;
		/* Keystrokes for extend-select shouldn't be able to dbl-click. */

	canGoSlow = false;
		/* Since we are using TEClick to handle the hilight, make sure that
		** our slow-select feature doesn't work for keys. */

	TEClick(fakeClick, extend, teHndl);
		/* Do new selection without flash like extend-select with mouse does it. */

	(*teHndl)->clickTime = 0L;
		/* We just did a key, not a click, so reset the click timer. */

	selBot = (fakeClick.v += (lineHeight >> 1));
	selTop = (fakeClick.v -= lineHeight);

	viewRct = (*teHndl)->viewRect;
	dy = 0;
	if (selTop < viewRct.top)    dy = selTop - viewRct.top;
	if (selBot > viewRct.bottom) dy = selBot - viewRct.bottom;

	--fakeClick.h;
	dx = 0;
	if (fakeClick.h < viewRct.left)  dx = fakeClick.h - viewRct.left;
	if (fakeClick.h > viewRct.right) dx = fakeClick.h - viewRct.right + 1;

	if (dx | dy) TEScroll(-dx, -dy, teHndl);

	TEAutoView(true, teHndl);
	(*teHndl)->clikLoop = ASMTECLIKLOOP;
	AdjustTEBottom(teHndl);
	AdjustScrollValues(teHndl);
	SetPort(oldPort);
}



/*****************************************************************************/



/* This is called when a mouseDown occurs in the content of a window.  It
** returns true if the mouseDown caused a TextEdit action to occur.  Events
** that are handled include if the user clicks on a scrollbar that is
** associated with a TextEdit control.
*/

#pragma segment Controls
Boolean	CTEClick(EventRecord *event)
{
	WindowPtr		oldPort, window;
	Point			mouseLoc;
	TEHandle		te, teActive;
	ControlHandle	ctlHit, viewCtl;
	CTEDataHndl		teData;
	Boolean			vert;
	short			extendSelect, part, value;

	GetPort(&oldPort);
	if (!(window = FrontWindow())) return(false);

	SetPort(window);
	mouseLoc = event->where;
	GlobalToLocal(&mouseLoc);

	if (CTEFindCtl(window, event, &te, &ctlHit)) {
			/* See if the user clicked directly on the view control for a
			** TextEdit record.  If so, we definitely have some work to do.
			*/
		if (te != gActiveTEHndl) {
			CTEActivate(true, te);
			SetPort(oldPort);
			return(true);
				/* If user clicked on TextEdit control other than the
				** currently active control, then activate it.  This is our
				** only action in this case.
				*/
		}
		extendSelect = ((event->modifiers & shiftKey) != 0);
			/* Extend-select may be occuring. */

		canGoSlow = true;
			/* There is a slow-zone around the TextEdit area for slow extend-select.
			** Allow this zone to slow down selection. */

		TEClick(mouseLoc, extendSelect, te);
			/* Do the extend-select thing.  Most of the work is handled by
			** TextEdit.  The only thing we have to do is to update the
			** scrollbars while the user is extending the select.  This is
			** taken care of by our custom clikLoop procedure.
			*/

		if (viewCtl = CTEViewFromTE(te)) {
			teData = (CTEDataHndl)(*viewCtl)->contrlData;
			(*teData)->newUndo = true;
		}

		SetPort(oldPort);
		return(true);
	}

/* We didn't hit the view control for a TextEdit record, but don't give up yet.
** The user may be clicking on a related scrollbar.  Let's find out...
*/

	if (part = FindControl(mouseLoc, window, &ctlHit)) {
			/* The user did click on a control.  But is it a scrollbar
			** for a TextEdit control?  Stay tuned...
			*/
		te = CTEFromScroll(ctlHit, &viewCtl);

		if (te) {		/* It was a related scrollbar. */

			if (te != gActiveTEHndl) {
				CTEActivate(true, te);
				SetPort(oldPort);
				return(true);
					/* If user clicked on TextEdit control other than the
					** currently active control, then activate it.  This is our
					** only action in this case.
					*/
			}

			CTEActivate(true, te);

			vert = ((*ctlHit)->contrlRect.top <= (*viewCtl)->contrlRect.top);
				/* Horizontal or vertical scroll.  Only the rect knows. */

			switch (part) {
				case inThumb:
					value = GetCtlValue(ctlHit);
					part = TrackControl(ctlHit, mouseLoc, nil);
					if (part) {
						value -= GetCtlValue(ctlHit);
							/* Value now has CHANGE in value.
							** if value changed, scroll. */
						if (value) {
							if (vert)
								TEScroll(0, value, te);
							else
								TEScroll(value, 0, te);
						}
					}
					break;

				default:
					teActive = gActiveTEHndl;
					gActiveTEHndl = te;
						/* This is a hack way to pass the action procedure
						** which TextEdit record we are dealing with.
						*/
					if (vert)
						TrackControl(ctlHit, mouseLoc, (ProcPtr)VActionProc);
					else
						TrackControl(ctlHit, mouseLoc, (ProcPtr)HActionProc);

					gActiveTEHndl = teActive;
						/* Unhack our previous hack. */
					break;
			}
			SetPort(oldPort);
			return(true);
		}
	}

	SetPort(oldPort);
	return(false);
}



/*****************************************************************************/



/* This is the custom clikLoop, which is called from the assembly glue code.
** This handles updating the scrollbars as the user is drag-selecting in
** the TextEdit control.
*/

#pragma segment Controls
void	CTEClikLoop(void)
{
	WindowPtr		oldPort, window;
	TEHandle		te;
	Point			mouseLoc;
	Rect			viewRct;
	RgnHandle		rgn;
	short			dl, dr, dt, db, lh;
	long			tick;

	GetPort(&oldPort);
	SetPort(window = (WindowPtr)(*gActiveTEHndl)->inPort);

	te = gActiveTEHndl;
		/* This better be what the user is dragging, or we did something
		** wrong elsewhere.
		*/

	GetMouse(&mouseLoc);
	viewRct = (*te)->viewRect;

	if (!PtInRect(mouseLoc, &viewRct)) {
		/* User is outside the TextEdit area, so scrolling is happening. */

		if (tick = canGoSlow) tick = TickCount();
			/* As an extra feature, there is a zone around the TextEdit
			** viewRect that the scroll will be slowed down.  This zone
			** is based on the lineHeight of the active TextEdit control.
			** If the user drags outside the viewRect further than the
			** value of lineHeight, then scrolling occurs as fast as possible.
			*/

		rgn = NewRgn();
		GetClip(rgn);
		ClipRect(&(window->portRect));
			/* The clipRgn is set to protect everything outside viewRect.
			** This doesn't work very well when we want to change
			** the scrollbars.  Save the old and open it up.
			*/

		dl = viewRct.left - mouseLoc.h;
		dr = mouseLoc.h   - viewRct.right;
		dt = viewRct.top  - mouseLoc.v;
		db = mouseLoc.v   - viewRct.bottom;
			/* Check the delta value for each side of viewRect.  This will
			** be used to determine if we should scroll fast or slow.
			*/

		AdjustScrollValues(te);
			/* Scroll them puppies. */

		SetClip(rgn);								/* restore clip */
		DisposeRgn(rgn);
			/* Make Mr. clipRgn happy again. */

		lh = (*te)->lineHeight;
		if ((dl < lh) && (dr < lh) && (dt < lh) && (db < lh))
			while (TickCount() <= tick + 9);
				/* Do it really slow.  (This is important on an fx!!) */
	}

	SetPort(oldPort);
}



/*****************************************************************************/



/* Do the cut/copy/paste/clear operations for the currently active
** TextEdit control.  Caller assumes appropriateness of the call.  Typically,
** this routine won't be called at an inappropriate time, since the menu
** item should be enabled or disabled correctly.
** Use CTEEditMenu to set the menu items undo/cut/copy/paste/clear correctly
** for the active TextEdit control.  Since undo isn't currently supported,
** all that CTEEditMenu does for the undo case is to deactivate it right now.
*/

#pragma segment Controls
void	CTEClipboard(short menuID)
{
	WindowPtr		oldPort;
	TEHandle		te;
	ControlHandle	viewCtl;
	short			maxTextLen, charsToAdd;

	if (!(te = gActiveTEHndl)) return;
	if (!(viewCtl = CTEViewFromTE(te))) return;

	GetPort(&oldPort);
	SetPort((*te)->inPort);
	switch (menuID) {
		case 2:
			CTENewUndo(viewCtl, true);
			TECut(te);
			ZeroScrap();
			if (TEToScrap()) ZeroScrap();
			break;
		case 3:
			TECopy(te);
			ZeroScrap();
			if (TEToScrap()) ZeroScrap();
			break;
		case 4:
			TEFromScrap();
			if (viewCtl) {
				maxTextLen = (*(CTEDataHndl)((*viewCtl)->contrlData))->maxTextLen;
				charsToAdd = TEGetScrapLen() - ((*te)->selEnd - (*te)->selStart);
				if ((*te)->teLength + charsToAdd <= maxTextLen) {
					CTENewUndo(viewCtl, true);
					TEPaste(te);
				}
			}
			break;
		case 5:
			CTENewUndo(viewCtl, true);
			TEDelete(te);
			break;
	}

	AdjustTEBottom(te);
	AdjustScrollValues(te);
	SetPort(oldPort);
}



/*****************************************************************************/



#pragma segment Controls
pascal long	CTECtl(short varCode, ControlHandle ctl, short msg, long parm)
{
#pragma unused (varCode)

	Rect			viewRct;
	TEHandle		te;
	CTEDataHndl		teData;
	ControlHandle	scrollCtl;
	short			i;

	if (te = (TEHandle)GetCRefCon(ctl)) viewRct = (*ctl)->contrlRect;
	else SetRect(&viewRct, 0, 0, 0, 0);

	switch (msg) {
		case drawCntl:
			CTEUpdate(te, ctl, false);
			break;

		case testCntl:
			if (PtInRect(*(Point *)&parm, &viewRct)) {
				gFoundViewCtl = ctl;
				gFoundTEHndl  = te;
				return(1);
			}
			return(0);
			break;

		case calcCRgns:
		case calcCntlRgn:
			if (msg == calcCRgns) parm &= 0x00FFFFFF;
			RectRgn((RgnHandle)parm, &viewRct);
			break;

		case initCntl:
			break;

		case dispCntl:
			if (te) {
				if (te == gActiveTEHndl) gActiveTEHndl = nil;
				TEDispose(te);
				if (teData = (CTEDataHndl)(*ctl)->contrlData) {
					if ((*teData)->undoText) DisposHandle((Handle)(*teData)->undoText);
					DisposHandle((Handle)teData);
				}
				for (i = 0; i < 2; ++i)
					if (scrollCtl = CTEScrollFromView(ctl, i))
						DisposeControl(scrollCtl);
			}
			break;

		case posCntl:
			break;

		case thumbCntl:
			break;

		case dragCntl:
			break;

		case autoTrack:
			break;
	}

	return(0);
}



/*****************************************************************************/



#pragma segment Controls
ControlHandle	CTECtlHit(void)
{
	ControlHandle	ctl;

	ctl = gFoundViewCtl;
	gFoundViewCtl = nil;
	return(ctl);
}



/*****************************************************************************/



#pragma segment Controls
void	CTEDispose(TEHandle teHndl)
{
	WindowPtr	oldPort;

	if (teHndl) {
		GetPort(&oldPort);
		SetPort((*teHndl)->inPort);
		TEDispose(CTEDisposeView(CTEViewFromTE(teHndl)));
			/* Dispose of the TextEdit control completely.  This includes
			** scrollbars, as well as the TextEdit view control.
			*/
		SetPort(oldPort);
	}
}



/*****************************************************************************/



/* Dispose of the view control and related scrollbars.  This function also
** returns the handle to the TextEdit record, since it was just orphaned.
** Use this function if you want to get rid of a TextEdit control, but you
** want to keep the TextEdit record.
*/

#pragma segment Controls
TEHandle	CTEDisposeView(ControlHandle viewCtl)
{
	TEHandle		te;
	short			vert;
	ControlHandle	ctl;

	if (!viewCtl) return(nil);

	te = (TEHandle)GetCRefCon(viewCtl);
	SetCRefCon(viewCtl, (long)nil);

	for (vert = 0; vert < 2; ++vert)
		if (ctl = CTEScrollFromView(viewCtl, vert))
			DisposeControl(ctl);

	DisposeControl(viewCtl);
	return(te);
}



/*****************************************************************************/



#pragma segment Controls
short	CTEDocHeight(TEHandle teHndl)
{
	if (!teHndl) return(0);

	return(CTENumTextLines(teHndl) * (*teHndl)->lineHeight);
}



/*****************************************************************************/



/* Enable or disable edit menu items based on the active TextEdit control.
** You pass the menu ID of the undo item in undoID, and the menu ID of the
** cut item in cutID.  If undoID or cutID is non-zero, then some action is
** performed.  Since I don't support TextEdit control undo yet (next version?),
** all that passing a non-zero value for undoID does is disable the undo
** menu item.  If you pass a non-zero value for cutID, then the other menu
** items cut/copy/paste/clear are updated to reflect the status of the
** active TextEdit control.
*/

#pragma segment Controls
Boolean	CTEEditMenu(Boolean *activeItem, short editMenu, short undoID, short cutID)
{
	TEHandle		te;
	MenuHandle		menu;
	Boolean			active;
	ControlHandle	viewCtl;
	CTEDataHndl		teData;

	*activeItem = active = false;
	menu = GetMHandle(editMenu);

	if (undoID)
		DisableItem(menu, undoID);
	if (cutID) {
		DisableItem(menu, cutID);			/* Disable cut. */
		DisableItem(menu, cutID + 1);		/* Disable copy. */
		DisableItem(menu, cutID + 2);		/* Disable paste. */
		DisableItem(menu, cutID + 3);		/* Disable clear. */
	}

	if (!(te = gActiveTEHndl)) return(false);

	if (undoID) {
		if (viewCtl = CTEViewFromTE(te)) {
			teData = (CTEDataHndl)(*viewCtl)->contrlData;
			if ((*teData)->undoText) {
				EnableItem(menu, undoID);
				active = true;
			}
		}
	}

	if (cutID) {
		if ((*te)->selStart != (*te)->selEnd) {
			if (!CTEReadOnly(te)) {
				EnableItem(menu, cutID);		/* Enable cut. */
				EnableItem(menu, cutID + 3);	/* Enable clear. */
			}
			active = true;
			EnableItem(menu, cutID + 1);		/* Enable copy. */
		}
		if (!CTEReadOnly(te)) {
			TEFromScrap();
			if (TEGetScrapLen()) {
				active = true;
				EnableItem(menu, cutID + 2);		/* Enable paste. */
			}
		}
	}

	*activeItem = active;
	return(true);
}



/*****************************************************************************/



/* Handle the event if it applies to the active TextEdit control.  If some
** action occured due to the event, return true.
*/

#pragma segment Controls
Boolean	CTEEvent(EventRecord *event)
{
	WindowPtr	window;

	switch(event->what) {

		case mouseDown:
			if (FindWindow(event->where, &window) == inContent)
				if (window == FrontWindow())
					return(CTEClick(event));
			break;

		case autoKey:
		case keyDown:
			if (!(event->modifiers & cmdKey))
				return(CTEKey(event));
			break;
	}

	return(false);
}



/*****************************************************************************/



/* This determines if a TextEdit control was clicked on directly.  This does
** not determine if a related scrollbar was clicked on.  If a TextEdit
** control was clicked on, then true is returned, as well as the TextEdit
** handle and the handle to the view control.
*/

#pragma segment Controls
Boolean	CTEFindCtl(WindowPtr window, EventRecord *event,
				   TEHandle *teHndl, ControlHandle *ctlHit)
{
	WindowPtr		oldPort;
	Point			mouseLoc;

	if (window) {
		GetPort(&oldPort);
		SetPort(window);
		mouseLoc = event->where;
		GlobalToLocal(&mouseLoc);
		SetPort(oldPort);

		gFoundTEHndl = nil;
		FindControl(mouseLoc, window, ctlHit);
		if (*teHndl = gFoundTEHndl) return(true);
	}

	*ctlHit = nil;
	return(false);
}



/*****************************************************************************/



#pragma segment Controls
TEHandle	CTEFindActive(WindowPtr window)
{
	if (!window) return(gActiveTEHndl);
		/* User wants whatever is active one, for whatever window. */

	if (!gActiveTEHndl) return(nil);
	if (window != (*gActiveTEHndl)->inPort) return(nil);
	return(gActiveTEHndl);
}



/*****************************************************************************/



/* Find the TextEdit record that is related to the indicated scrollbar. */

#pragma segment Controls
TEHandle	CTEFromScroll(ControlHandle scrollCtl, ControlHandle *retCtl)
{
	WindowPtr		window;
	ControlHandle	viewCtl;
	TEHandle		te;

	if (!IsScrollBar(scrollCtl)) {
		*retCtl = nil;
		return(nil);
	}

	window = (*scrollCtl)->contrlOwner;

	for (*retCtl = viewCtl = nil;;) {
		viewCtl = CTENext(window, &te, viewCtl);
		if (!viewCtl) return(nil);
		if (viewCtl == (ControlHandle)GetCRefCon(scrollCtl)) {
			*retCtl = viewCtl;
			return(te);
		}
	}
}



/*****************************************************************************/



#pragma segment Controls
void	CTEHide(TEHandle teHndl)
{
	ControlHandle	viewCtl, scrollCtl;
	short			i;

	if (teHndl) {
		CTEActivate(false, teHndl);
		viewCtl = CTEViewFromTE(teHndl);
		if (viewCtl) {
			HideControl(viewCtl);
			for (i = 0; i < 2; i++) {
				scrollCtl = CTEScrollFromView(viewCtl, i);
				if (scrollCtl) HideControl(scrollCtl);
			}
		}
	}
}



/*****************************************************************************/



/* Blink the caret in the active TextEdit control.  The active TextEdit
** control may be read-only, in which case the caret does not blink. */

#pragma segment Controls
void	CTEIdle(void)
{
	WindowPtr	window;

	if (gActiveTEHndl)
		if (window = FrontWindow())
			if (((WindowPeek)window)->windowKind >= userKind)
				TEIdle(gActiveTEHndl);
}



/*****************************************************************************/



/* See if the keypress event applies to the TextEdit control, and if it does,
** handle it and return true.
*/

#pragma segment Controls
short	CTEKey(EventRecord *event)
{
	TEHandle		te;
	ControlHandle	viewCtl;
	short			maxTextLen, selStart, selEnd;
	short			textSelected, arrowKey, lineHeight, retval;
	char			key;
	CTEDataHndl		teData;
	Point			pt;
#ifdef THINK_C
	long			tempLong;
#endif

	if (!(te = gActiveTEHndl))          return(0);
	if (CTEReadOnly(te))                return(0);
	if (!(viewCtl = CTEViewFromTE(te))) return(0);

	teData     = (CTEDataHndl)(*viewCtl)->contrlData;
	maxTextLen = (*teData)->maxTextLen;
	key        = event->message & charCodeMask;
	selStart   = (*te)->selStart;
	selEnd     = (*te)->selEnd;
	if (selStart > selEnd) {
		selStart = selEnd;
		selEnd = (*te)->selStart;
	}
	textSelected = (selStart != selEnd);
	arrowKey     = ((key >= kLeftArrow) && (key <= kDownArrow));
	lineHeight   = (*te)->lineHeight;

	retval = 1;
	if (
		(textSelected) ||				/* If selection range to be replaced or */
		(arrowKey) ||					/* key is an arrow or					*/
		(key == 8) ||					/* key is a delete or                   */
		((*te)->teLength < maxTextLen)	/* we have space for the key...         */
	) {
		CTENewUndo(viewCtl, false);		/* Add the character. */
		if (arrowKey) {
			if (event->modifiers & shiftKey) {
				switch (key) {
					case kUpArrow:
#ifndef THINK_PRE_5
						pt       = TEGetPoint(selStart, te);
#else
						tempLong = TEGetPoint(selStart, te);
						BlockMove(&tempLong, &pt, 4);
#endif
						pt.v -= (lineHeight >> 1);
						if (selStart == TEGetOffset(pt, te)) pt.v -= lineHeight;
						selStart = TEGetOffset(pt, te);
						CTEFakeClick(selStart, selEnd, true, te);
						break;
					case kDownArrow:
#ifndef THINK_PRE_5
						pt       = TEGetPoint(selEnd, te);
#else
						tempLong = TEGetPoint(selEnd, te);
						BlockMove(&tempLong, &pt, 4);
#endif
						selEnd = TEGetOffset(pt, te);
						CTEFakeClick(selStart, selEnd, true, te);
						break;
					case kLeftArrow:
						CTEFakeClick(--selStart, selEnd, true, te);
						break;
					case kRightArrow:
						CTEFakeClick(selStart, ++selEnd, true, te);
						break;
				}
				AdjustTEBottom(te);
				AdjustScrollValues(te);
				return(retval);
			}
			if (textSelected) {
				switch (key) {
					case kUpArrow:
					case kLeftArrow:
						TESetSelect(selStart, selStart, te);
						if (key == kUpArrow) textSelected = false;
						break;
					case kRightArrow:
					case kDownArrow:
						TESetSelect(selEnd, selEnd, te);
						if (key == kDownArrow) textSelected = false;
						break;
				}
				if (textSelected) {
					AdjustTEBottom(te);
					AdjustScrollValues(te);
					return(retval);
				}
			}
		}

		switch (key) {
			case kUpArrow:
#ifndef THINK_PRE_5
				pt       = TEGetPoint(selStart, te);
#else
				tempLong = TEGetPoint(selStart, te);
				BlockMove(&tempLong, &pt, 4);
#endif
				pt.v -= (lineHeight >> 1);
				if (selStart == TEGetOffset(pt, te)) pt.v -= lineHeight;
				selStart = TEGetOffset(pt, te);
				CTEFakeClick(selStart, selStart, false, te);
				break;
			case kLeftArrow:
				CTEFakeClick(--selStart, selEnd, false, te);
				break;
			case kRightArrow:
				CTEFakeClick(selStart, ++selEnd, false, te);
				break;
			default:
				TEKey(key, te);
				if (key != kDownArrow) ++retval;
				break;
		}

		AdjustTEBottom(te);
		AdjustScrollValues(te);
	}

	return(retval);
}



/*****************************************************************************/



/* This function is used to move a TextEdit control.  Pass it the TextEdit
** record to move, plus the new position.  It will move the TextEdit control,
** along with any scrollbars the control may have.  All areas that need
** updating are cleared and invalidated.
*/

#pragma segment Controls
void	CTEMove(TEHandle teHndl, short newH, short newV)
{
	WindowPtr		oldPort;
	Rect			viewRct, brdrRct, rct;
	short			i, dx, dy, mode;
	ControlHandle	viewCtl, ctl;
	CTEDataHndl		teData;

	if (!(viewCtl = CTEViewFromTE(teHndl))) return;

	GetPort(&oldPort);
	SetPort((*teHndl)->inPort);
	teData = (CTEDataHndl)(*viewCtl)->contrlData;

	viewRct = (*viewCtl)->contrlRect;
	EraseRect(&viewRct);
	InvalRect(&rct);
	brdrRct = (*teData)->brdrRect;
	EraseRect(&brdrRct);
	InvalRect(&brdrRct);

	dx = newH - viewRct.left;
	dy = newV - viewRct.top;

	for (i = 0; i < 2; ++i) {
		if (ctl = CTEScrollFromView(viewCtl, i)) {
			rct = (*ctl)->contrlRect;
			MoveControl(ctl, rct.left + dx, rct.top + dy);
		}
	}

	OffsetRect(&(*viewCtl)->contrlRect, dx, dy);
	OffsetRect(&(*teHndl)->destRect, dx, dy);
	OffsetRect(&(*teHndl)->viewRect, dx, dy);
	OffsetRect(&(*teData)->brdrRect, dx, dy);

	rct = viewRct;
	OffsetRect(&rct, dx, dy);

	mode = (*teData)->mode;
	rct.top  = rct.bottom - 16;
	rct.left = rct.right - 16;

	if (mode & (cteVScrollLessGrow - cteVScroll + cteHScrollLessGrow - cteHScroll))
		EraseRect(&rct);

	if (mode & (cteVScrollLessGrow - cteVScroll)) OffsetRect(&rct, 16, 0);
	if (mode & (cteHScrollLessGrow - cteHScroll)) OffsetRect(&rct, 0, 16);
	InvalRect(&rct);

	SetPort(oldPort);
}



/*****************************************************************************/



/* Create a new TextEdit control.  See the comments at the beginning of this
** file for more information.
*/

#pragma segment Controls
void	CTENew(short viewID, WindowPtr window, TEHandle *teHndl, Rect *cRect, Rect *dRect,
			   Rect *vRect, Rect *bRect, short maxTextLen, short mode)
{
	Rect			destRect, viewRect, brdrRect;
	WindowPtr		oldPort;
	TEHandle		te;
	Boolean			err;
	short			width, height;
	Rect			ctlRect;
	ControlHandle	viewCtl, hScrollCtl, vScrollCtl;
	cdefRsrcJMPHndl	cdefRsrc;
	CTEDataHndl		teData;

	theViewID = viewID;		/* Keep viewID that was passed in. */

	GetPort(&oldPort);
	SetPort(window);

	destRect = *dRect;
	viewRect = *vRect;
	brdrRect = *bRect;
		/* Make sure that the rects are not in memory that may move. */

	te = TENew(&destRect, &viewRect);
		/* Do the main thing. */

	err = false;
	viewCtl = hScrollCtl = vScrollCtl = nil;
		/* Prepare for various failures. */

	if (te) {		/* If we were able to create the TextEdit record... */

		TEAutoView(true, te);
			/* Let TextEdit 3.0 do most of the scrolling work. */
		gDefaultClikLoop = (*te)->clikLoop;
		(*te)->clikLoop  = ASMTECLIKLOOP;
			/* We will do the remainder of the work. */

		cdefRsrc = (cdefRsrcJMPHndl)GetResource('CDEF', viewID);
		(*cdefRsrc)->jmpAddress = (long)CTECtl;
		FlushInstructionCache();
			/* Make sure that instruction caches don't kill us. */

		viewCtl = NewControl(window, cRect, nil, false, 0, 0, 0,
							 viewID * 16, (long)te);
			/* Use our custom view cdef.  It's wierd, but it's small. */

		if (!viewCtl) err = true;
		else {
			(*viewCtl)->contrlData = nil;
			if (teData = (CTEDataHndl)NewHandle(sizeof(CTEDataRec))) {
				(*teData)->maxTextLen  = maxTextLen;
				(*teData)->undoText    = nil;
				(*teData)->mode        = mode;
				(*teData)->brdrRect    = brdrRect;
				(*viewCtl)->contrlData = (Handle)teData;
			}
			else err = true;

			if (mode & cteHScroll) {		/* Caller wants a horizontal scrollbar... */
				SetRect(&ctlRect, 0, 0, 100, 16);
				hScrollCtl = NewControl(window, &ctlRect, nil, true, 0, 0, 0,
										scrollBarProc, (long)viewCtl);
				if (hScrollCtl) {
					MoveControl(hScrollCtl, brdrRect.left, brdrRect.bottom - 1);
					width = brdrRect.right - brdrRect.left;
					if (mode & (cteHScrollLessGrow - cteHScroll))
						if (!(mode & cteVScroll)) width -= 15;
					SizeControl(hScrollCtl, width, 16);
						/* Line the scrollbar up with the borderRect. */
				}
				else err = true;
			}

			if (mode & cteVScroll) {		/* Caller wants a vertical scrollbar... */
				SetRect(&ctlRect, 0, 0, 16, 100);
				vScrollCtl = NewControl(window, &ctlRect, nil, true, 0, 0, 0,
										scrollBarProc, (long)viewCtl);
				if (vScrollCtl) {
					MoveControl(vScrollCtl, brdrRect.right - 1, brdrRect.top);
					height = brdrRect.bottom - brdrRect.top;
					if (mode & (cteVScrollLessGrow - cteVScroll))
						if (!(mode & cteHScroll)) height -= 15;
					SizeControl(vScrollCtl, 16, height);
						/* Line the scrollbar up with the borderRect. */
				}
				else err = true;
			}
		}
	}
	else err = true;

	SetPort(oldPort);

	if (err) {		/* Oops.  Somebody wasn't happy. */
		if (viewCtl)
			DisposeControl(viewCtl);
				/* This also disposes of TextEdit handle! */
		else
			if (te) TEDispose(te);
				/* We have to dispose of the TextEdit handle ourselves if
				** creating the view control failed. */

		te = nil;		/* Return that there is no TextEdit control. */

		if (hScrollCtl)
			DisposeControl(hScrollCtl);
				/* More clean-up. */

		if (vScrollCtl)
			DisposeControl(vScrollCtl);
				/* And still more clean-up. */

		*teHndl = nil;		/* Return that we failed. */
	}
	else {
		ShowControl(viewCtl);		/* Since everything worked, show the control. */

		if (mode & cteReadOnly)
			(*te)->caretHook = (ProcPtr)ASMNOCARET;
				/* If read-only, then disable caret for this
				** TextEdit control.
				*/

		if (mode & cteActive) CTEActivate(true, te);

		AdjustScrollValues(*teHndl = te);
			/* Give the scrollbars an initial value.  This is because the
			** TextEdit control could have been created with
			** destRect.top < viewRect.top or destRect.left < viewRect.left.
			** I don't know why anyone would want to, but it is legal.
			*/
	}
}



/*****************************************************************************/



/* Save the data (if appropriate) so that user can undo. */

#pragma segment Controls
void	CTENewUndo(ControlHandle viewCtl, Boolean alwaysNewUndo)
{
	TEHandle	teHndl;
	CTEDataHndl	teData;
	Handle		hText, uText;

	if (!viewCtl) return;

	teHndl = (TEHandle)GetCRefCon(viewCtl);
	teData = (CTEDataHndl)(*viewCtl)->contrlData;

	hText = (*teHndl)->hText;
	uText = (*teData)->undoText;
		/* hText is text in the TextEdit record that is being edited.  */
		/* uText is the undo data (if any) prior to current edit task. */

	if (!uText) {				/* No undo text (therefore no editing) yet. */
		uText = NewHandle((*teHndl)->teLength);
		if (!uText) return;		/* Not enough memory to support undo. */
		(*teData)->undoText = uText;
		alwaysNewUndo = true;
			/* Since this is our first edit, we will want to cache the
			** data, no matter for what reason we were called. */
	}

	if ((alwaysNewUndo) || ((*teData)->newUndo)) {
		SetHandleSize(uText, (*teHndl)->teLength);
		if (MemError()) return;
			/* Not enough memory to handle this undo. */
		BlockMove(*hText, *uText, (*teHndl)->teLength);
		(*teData)->newUndo      = false;
		(*teData)->undoSelStart = (*teHndl)->selStart;
		(*teData)->undoSelEnd   = (*teHndl)->selEnd;
	}
}



/*****************************************************************************/



/* Get the next TextEdit control in the window.  You pass it a control handle
** for the view control, or nil to start at the beginning of the window.
** It returns both a TextEdit handle and the view control handle for that
** TextEdit record.  If none is found, nil is returned.  This allows you to
** repeatedly call this function and walk through all the TextEdit controls
** in a window.
*/

#pragma segment Controls
ControlHandle	CTENext(WindowPtr window, TEHandle *teHndl, ControlHandle ctl)
{
	short	defProcID;
	ResType	defProcType;
	Str255	defProcName;

	*teHndl = nil;
	if (!window) return(nil);

	if (!ctl)
		ctl = ((WindowPeek)window)->controlList;
	else
		ctl = (*ctl)->nextControl;

	while (ctl) {
		defProcID = !theViewID;
		GetResInfo((*ctl)->contrlDefProc, &defProcID, &defProcType, defProcName);
		if (defProcID == theViewID) {
			*teHndl = (TEHandle)GetCRefCon(ctl);
			break;
		}
		ctl = (*ctl)->nextControl;
	}

	return(ctl);
}



/*****************************************************************************/



/* Return the number of lines of text.  This is because there is a bug in
** TextEdit where the number of lines returned is incorrect if the text
** ends with a c/r.  This function adjusts for this bug.
*/

#pragma segment Controls
short	CTENumTextLines(TEHandle teHndl)
{
	short	lines;
	char	*cptr;

	if (!teHndl) return(0);

	lines = (*teHndl)->nLines;

	cptr = *((*teHndl)->hText);		/* Pointer to first TextEdit character. */
	if (cptr[(*teHndl)->teLength - 1] == kCrChar) ++lines;
		/* Since nLines isnt right if the last character is a return,
		** check for that case and fix it.
		*/

	return(lines);
}



/*****************************************************************************/



/* Return the number of text lines in the view area. */

#pragma segment Controls
short	CTENumViewLines(TEHandle teHndl)
{
	short	viewHeight;

	if (!teHndl) return(0);

	viewHeight = (*teHndl)->viewRect.bottom - (*teHndl)->viewRect.top;
	return(viewHeight / (*teHndl)->lineHeight);
}



/*****************************************************************************/



/* Use this function to print the contents of a TextEdit record.  Pass it a
** TextEdit handle, a pointer to a text offset, and a pointer to a rect to
** print the text in.  The offset should be initialized to what character
** in the TextEdit record you wish to start printing at (most likely 0).
** The print function prints as much text as will fit in the rect, and
** then updates the offset to tell you what is the first character that didn't
** print.  You can then call the print function again with another rect with
** this new offset, and it will print the text starting at the new offset.
** This method is very useful when a single TextEdit record is longer than a
** single page, and you wish the text to break at the end of the page.
** The bottom of rect is also updated, along with the offset.  The bottom edge
** of the rect is changed to reflect the actual bottom of the text printed.
** This is useful because the rect passed in didn't necessarily hold an
** integer number of lines of text.  The bottom of the rect is adjusted so
** it exactly holds complete lines of text.
** It is also possible that the rect could hold substantially more lines of
** text than there are remaining.  Again, in this situation, the bottom of
** rect is adjusted so that the rect tightly bounds the text printed.
** The remaining piece of information passed back is an indicator that the
** text through the end of the TextEdit record was printed.  When the end
** of the text is reached, the offset for the next text to be printed is
** returned as -1.  This indicates that processing of the TextEdit record
** is complete.
*/

#pragma segment Controls
OSErr	CTEPrint(TEHandle teHndl, short *teOffset, Rect *teRct)
{
	short		len, offset, numLines, rctHeight, rctLines, h;
	Handle		hText, keepHText;
	Rect		keepDestRect, keepViewRect;
	WindowPtr	tePort, printPort;

	if (teHndl) return(noErr);

	len = (*teHndl)->teLength;
	if (!(keepHText = NewHandle(len))) return(memFullErr);

	if ((offset = *teOffset) >= len) {
		*teOffset = -1;					/* We are offset further than we have text. */
		teRct->bottom = teRct->top;		/* Empty rect, since no text to draw.  */
		return(noErr);					/* Just say that we have no more text. */
	}

	BlockMove(*(hText = (*teHndl)->hText), *keepHText, len);
	keepDestRect = (*teHndl)->destRect;
	keepViewRect = (*teHndl)->viewRect;
	tePort = (*teHndl)->inPort;
		/* Cache some information from the TextEdit record. */

	BlockMove(*hText + offset, *hText, len - offset);
	SetHandleSize(hText, (*teHndl)->teLength = len - offset);
		/* Throw out the characters that have already been printed. */

	GetPort(&printPort);
	(*teHndl)->inPort = printPort;
	(*teHndl)->destRect = (*teHndl)->viewRect = *teRct;
	TECalText(teHndl);
		/* Install the print rect into the TextEdit record and then
		** rewrap the unprinted text into this rectangle.  The text
		** is now formatted correctly to print this rect's worth. */

	numLines  = CTENumTextLines(teHndl);
	rctHeight = teRct->bottom - teRct->top;
	rctLines  = rctHeight / (h = (*teHndl)->lineHeight);

	if (rctLines > numLines) rctLines = numLines;
	teRct->bottom = teRct->top + rctLines * h;
	(*teHndl)->destRect = (*teHndl)->viewRect = *teRct;
		/* We now have the minimum rectangle to hold as much of the text
		** as would fit into the rectangle passed in.  The final calc on
		** the rect prevents partial lines from being drawn. */

	TEUpdate(teRct, teHndl);		/* Draw this portion of the text. */

	if (rctLines == numLines)
		*teOffset = -1;
			/* Printed the last of the text for this record. */
	else
		*teOffset = (*teHndl)->lineStarts[rctLines] + offset;
			/* Offset to the first character not printed. */

	SetHandleSize(hText, len);
	BlockMove(*keepHText, *hText, len);
	DisposHandle(keepHText);
	(*teHndl)->teLength = len;
	(*teHndl)->inPort   = tePort;
	(*teHndl)->destRect = keepDestRect;
	(*teHndl)->viewRect = keepViewRect;
	TECalText(teHndl);		/* Restore the TextEdit record to the way it was. */
	return(noErr);			/* Everything worked. */
}



/*****************************************************************************/



/* Return if the TextEdit control is read/write (true) or read-only (false). */

#pragma segment Controls
Boolean	CTEReadOnly(TEHandle teHndl)
{
	if (!teHndl) return(false);
	if ((*teHndl)->caretHook == ASMNOCARET) return(true);
	return(false);
}



/*****************************************************************************/



/* Return the control handle for the TextEdit control's scrollbar, either
** vertical or horizontal.  If the scrollbar doesn't, nil is returned.
*/

#pragma segment Controls
ControlHandle	CTEScrollFromTE(TEHandle teHndl, Boolean vertScroll)
{
	ControlHandle	viewCtl;

	if (!(viewCtl = CTEViewFromTE(teHndl))) return(nil);

	return(CTEScrollFromView(viewCtl, vertScroll));
}



/*****************************************************************************/



/* Return the control handle for the scrollbar related to the view control,
** either horizontal or vertical.  If the scrollbar doesn't exist, return nil.
*/

#pragma segment Controls
ControlHandle	CTEScrollFromView(ControlHandle viewCtl, Boolean vertScroll)
{
	ControlHandle	ctl;
	WindowPtr		window;
	Boolean			vert;

	if (!viewCtl) return(nil);

	window = (*viewCtl)->contrlOwner;
	ctl    = ((WindowPeek)window)->controlList;

	for (; ctl;) {
		if ((ControlHandle)GetCRefCon(ctl) == viewCtl) {
			vert = false;
			if ((*ctl)->contrlRect.right == (*ctl)->contrlRect.left + 16)
				vert = true;
			if (vert == vertScroll) return(ctl);
		}
		ctl = (*ctl)->nextControl;
	}
	return(nil);
}



/*****************************************************************************/



/* Select a range of text.  TESetSelect can't be used alone because it doesn't
** update the scrollbars.  This function calls TESetSelect, and then fixes up
** the scrollbars.
*/

#pragma segment Controls
void	CTESetSelect(short start, short end, TEHandle teHndl)
{
	WindowPtr		oldPort;
	ControlHandle	viewCtl;
	CTEDataHndl		teData;

	if (teHndl) {
		GetPort(&oldPort);
		SetPort((*teHndl)->inPort);
		TESetSelect(start, end, teHndl);
		AdjustScrollValues(teHndl);
		if (viewCtl = CTEViewFromTE(teHndl)) {
			teData = (CTEDataHndl)(*viewCtl)->contrlData;
			(*teData)->newUndo = true;
		}
		SetPort(oldPort);
	}
}



/*****************************************************************************/



#pragma segment Controls
void	CTEShow(TEHandle teHndl)
{
	ControlHandle	viewCtl, scrollCtl;
	short			i;

	if (viewCtl = CTEViewFromTE(teHndl)) {
		ShowControl(viewCtl);
		for (i = 0; i < 2; i++) {
			scrollCtl = CTEScrollFromView(viewCtl, i);
			if (scrollCtl) ShowControl(scrollCtl);
		}
	}
}



/*****************************************************************************/



/* This function is used to resize a TextEdit control.  Pass it the TextEdit
** record to resize, plus the new horizontal and vertical size.  It will
** resize the TextEdit control, realign the text, if necessary, plus it will
** resize and adjust any scrollbars the TextEdit control may have.  All areas
** that need updating are cleared and invalidated.
*/

#pragma segment Controls
void	CTESize(TEHandle teHndl, short newH, short newV, Boolean newDest)
{
	WindowPtr		oldPort;
	Rect			viewRct, brdrRct, teViewRct, oldTeViewRct, rct;
	short			i, dx, dy, mode;
	ControlHandle	viewCtl, ctl[2];
	CTEDataHndl		teData;
	RgnHandle		rgn1, rgn2;

	if (!(viewCtl = CTEViewFromTE(teHndl))) return;

	GetPort(&oldPort);
	SetPort((*teHndl)->inPort);

	viewRct = (*viewCtl)->contrlRect;
	teData  = (CTEDataHndl)(*viewCtl)->contrlData;
	brdrRct = (*teData)->brdrRect;
	mode    = (*teData)->mode;

	if (mode & (cteVScrollLessGrow - cteVScroll + cteHScrollLessGrow - cteHScroll)) {
		rct = viewRct;
		rct.top  = rct.bottom - 16;
		rct.left = rct.right - 16;
		if (mode & (cteVScrollLessGrow - cteVScroll)) OffsetRect(&rct, 16, 0);
		if (mode & (cteHScrollLessGrow - cteHScroll)) OffsetRect(&rct, 0, 16);
		EraseRect(&rct);
		InvalRect(&rct);
	}		/* Erase the old grow box, if we have one. */

	dx = newH - (viewRct.right  - viewRct.left);
	dy = newV - (viewRct.bottom - viewRct.top);

	for (i = 0; i < 2; ++i) {
		if (ctl[i] = CTEScrollFromView(viewCtl, i)) {
			rct = (*ctl[i])->contrlRect;
			if (i) {
				HideControl(ctl[i]);
				SizeControl(ctl[i], rct.right - rct.left, rct.bottom - rct.top + dy);
				MoveControl(ctl[i], rct.left + dx, rct.top);
			}
			else {
				HideControl(ctl[i]);
				SizeControl(ctl[i], rct.right - rct.left + dx, rct.bottom - rct.top);
				MoveControl(ctl[i], rct.left, rct.top + dy);
			}
		}
	}		/* Reposition the scrollbars, if we have any. */

	InvalRect(&viewRct);		/* Resize our view control. */
	viewRct.right  += dx;
	viewRct.bottom += dy;
	(*viewCtl)->contrlRect = viewRct;
	InvalRect(&viewRct);

	InvalRect(&brdrRct);		/* Resize our border rect. */
	brdrRct.right  += dx;
	brdrRct.bottom += dy;
	(*teData)->brdrRect = brdrRct;
	InvalRect(&brdrRct);

	teViewRct = oldTeViewRct = (*teHndl)->viewRect;		/* Resize TextEdit viewRect. */
	teViewRct.right  += dx;
	teViewRct.bottom += dy;
	(*teHndl)->viewRect = teViewRct;

	if (newDest) {		/* Resize TextEdit destRect, if so indicated. */
		(*teHndl)->destRect.right  += dx;
		(*teHndl)->destRect.bottom += dy;
		TECalText(teHndl);
		EraseRect(&oldTeViewRct);		/* Redraw the whole thing if destRect changes. */
		InvalRect(&oldTeViewRct);
	}

	rgn1 = NewRgn();
	rgn2 = NewRgn();
	RectRgn(rgn1, &oldTeViewRct);		/* Clear old areas if we are shrinking. */
	RectRgn(rgn2, &teViewRct);
	DiffRgn(rgn1, rgn2, rgn2);
	EraseRgn(rgn2);
	InvalRgn(rgn2);
	RectRgn(rgn2, &teViewRct);			/* Update new areas if we are growing. */
	DiffRgn(rgn2, rgn1, rgn2);
	InvalRgn(rgn2);
	DisposeRgn(rgn1);
	DisposeRgn(rgn2);

	for (i = 0; i < 2; ++i) if (ctl[i]) ShowControl(ctl[i]);
		/* Show the controls in their new position. */

	if (mode & (cteVScrollLessGrow - cteVScroll + cteHScrollLessGrow - cteHScroll)) {
		rct = viewRct;
		rct.top  = rct.bottom - 16;
		rct.left = rct.right - 16;
		if (mode & (cteVScrollLessGrow - cteVScroll)) OffsetRect(&rct, 16, 0);
		if (mode & (cteHScrollLessGrow - cteHScroll)) OffsetRect(&rct, 0, 16);
		EraseRect(&rct);
		InvalRect(&rct);
	}

	AdjustTEBottom(teHndl);
	AdjustScrollValues(teHndl);

	SetPort(oldPort);
}



/*****************************************************************************/



/* Swap the TextEdit text handle with the text handle passed in. */

#pragma segment Controls
Handle	CTESwapText(TEHandle teHndl, Handle newText, Boolean update)
{
	WindowPtr	oldPort;
	Handle		hText;
	Rect		destRect, viewRect;
	short		oldTextLen, newTextLen;

	if (!teHndl) return(nil);

	GetPort(&oldPort);
	SetPort((*teHndl)->inPort);

	hText = (*teHndl)->hText;

	newTextLen = GetHandleSize(newText);
	oldTextLen = (*teHndl)->teLength;
	SetHandleSize(hText, oldTextLen);
		/* Just to make sure TE is behaving itself. */

	(*teHndl)->hText = newText;
	(*teHndl)->teLength = newTextLen;

	TECalText(teHndl);

	if (update) {
		destRect = (*teHndl)->destRect;
		viewRect = (*teHndl)->viewRect;

		OffsetRect(&destRect,
			viewRect.left - destRect.left,
			viewRect.top  - destRect.top);

		(*teHndl)->destRect = destRect;
		(*teHndl)->selStart = (*teHndl)->selEnd = 0;

		EraseRect(&viewRect);
		TEUpdate(&viewRect, teHndl);
	}

	AdjustScrollValues(teHndl);

	SetPort(oldPort);
	return(hText);
}



/*****************************************************************************/



/* Return information for the currently active TextEdit control.  The currently
** active TextEdit control is stored in gActiveTEHndl, and can be accessed
** directly.  If gActiveTEHndl is nil, then there is no currently active one.
** The information that we return is the viewRect and window of the active
** TextEdit control.  This is information that could be gotten directly, but
** this call makes it a little more convenient.
*/

#pragma segment Controls
WindowPtr	CTETargetInfo(TEHandle *teHndl, Rect *teView)
{
	SetRect(teView, 0, 0, 0, 0);
	if (!(*teHndl = gActiveTEHndl)) return(nil);

	*teView = (*gActiveTEHndl)->viewRect;
	return((*gActiveTEHndl)->inPort);
}



/*****************************************************************************/



#pragma segment Controls
void	CTEUndo(void)
{
	TEHandle		teHndl;
	ControlHandle	viewCtl;
	CTEDataHndl		teData;
	Handle			hText, uText;
	short			oldStart, oldEnd;

	if (!(teHndl = gActiveTEHndl)) return;

	if (viewCtl = CTEViewFromTE(teHndl)) {
		teData = (CTEDataHndl)(*viewCtl)->contrlData;
		hText  = (*teHndl)->hText;
		uText  = (*teData)->undoText;
		if (!uText) return;		/* There is no undo yet.  Getting called in
								** this state is actually an error, but we
								** check, just in case.
								*/
		oldStart = (*teHndl)->selStart;
		oldEnd   = (*teHndl)->selEnd;
		(*teData)->undoText = CTESwapText(teHndl, (*teData)->undoText, false);
		(*teHndl)->selStart = (*teHndl)->selEnd = 0;
		CTEUpdate(teHndl, viewCtl, false);
		CTESetSelect((*teData)->undoSelStart, (*teData)->undoSelEnd, teHndl);
			/* CTESetSelect flags it as a new undo situation for us. */
		(*teData)->undoSelStart = oldStart;
		(*teData)->undoSelEnd   = oldEnd;
	}
}



/*****************************************************************************/



/* Draw the TextEdit control and frame. */

#pragma segment Controls
void	CTEUpdate(TEHandle teHndl, ControlHandle ctl, Boolean justShowActive)
{
	WindowPtr	oldPort, tePort;
	Rect		viewRect, brdrRect;
	CTEDataHndl	teData;
	short		mode;

	if (teHndl) {
		if ((*ctl)->contrlVis) {

			GetPort(&oldPort);
			SetPort(tePort = (*teHndl)->inPort);

			teData = (CTEDataHndl)(*ctl)->contrlData;
			viewRect = (*teHndl)->viewRect;
			brdrRect = (*teData)->brdrRect;
			FrameRect(&brdrRect);
	
			if (!justShowActive) {
				EraseRect(&viewRect);
				TEUpdate(&viewRect, teHndl);
			}
	
			mode = (*teData)->mode;
			if (mode & cteShowActive) {
				InsetRect(&brdrRect, -3, -3);
				if (mode & cteVScroll) brdrRect.right  += 15;
				if (mode & cteHScroll) brdrRect.bottom += 15;
				if ((tePort != FrontWindow()) || (teHndl != gActiveTEHndl)) PenPat(qd.white);
				PenSize(2, 2);
				FrameRect(&brdrRect);
				PenNormal();
			}

			SetPort(oldPort);
		}
	}
}



/*****************************************************************************/



/* Return the control handle for the view control that owns the TextEdit
** record.  Use this to find the view to do customizations such as changing
** the update procedure for this TextEdit control.
*/

#pragma segment Controls
ControlHandle	CTEViewFromTE(TEHandle teHndl)
{
	WindowPtr		window;
	ControlHandle	viewCtl;
	TEHandle		te;

	if (!teHndl) return(nil);

	window = (WindowPtr)(*teHndl)->inPort;

	for (viewCtl = nil;;) {

		viewCtl = CTENext(window, &te, viewCtl);
		if ((!viewCtl) || (te == teHndl)) return(viewCtl);
	}
}



/*****************************************************************************/



/* Call this when a window with TextEdit controls is being activated.  This
** will make the TextEdit control that was last active in this window the
** active TextEdit control again.  If it can't find one that was the last
** active control, it makes the first one it finds the active control.
*/

#pragma segment Controls
Boolean	CTEWindActivate(WindowPtr window)
{
	short			hilite, scrollNum;
	Boolean			oneIsActive;
	ControlHandle	viewCtl, scrollCtl;
	TEHandle		te;
	CTEDataHndl		teData;

	if (!window) return(false);

	hilite = 255;
	if (window == FrontWindow()) hilite = 0;

	for (oneIsActive = false, viewCtl = nil;;) {

		viewCtl = CTENext(window, &te, viewCtl);
		if (!viewCtl) break;
			/* The only way out of the loop. */

		teData = (CTEDataHndl)(*viewCtl)->contrlData;
		if ((*teData)->mode & cteActive) {
			if (!hilite) {
				CTEActivate(true, te);
				oneIsActive = true;
			}
			else CTEActivate(false, te);
		}

		for (scrollNum = 0; scrollNum < 2; ++scrollNum) {
			scrollCtl = CTEScrollFromTE(te, scrollNum);
			if (scrollCtl)
				HiliteControl(scrollCtl, hilite);
		}
	}

	return(oneIsActive);
}



/*****************************************************************************/
/*****************************************************************************/



#pragma segment Controls
pascal void	VActionProc(ControlHandle scrollCtl, short part)
{
	short		lineHeight, delta, value, teOffset;
	short		oldValue, max;
	TEHandle	te;
	
	if (part) {						/* If it was actually in the control. */

		te = gActiveTEHndl;
		lineHeight = (*te)->lineHeight;

		switch (part) {
			case inUpButton:
			case inDownButton:		/* One line. */
				delta = lineHeight;
				break;
			case inPageUp:			/* One page. */
			case inPageDown:
				delta = (*te)->viewRect.bottom - (*te)->viewRect.top;
				if (delta > lineHeight) delta -= lineHeight;
				break;
		}
		if ( (part == inUpButton) || (part == inPageUp) )
			delta = -delta;		/* Reverse direction for an upper. */

		value = (oldValue = GetCtlValue(scrollCtl)) + delta;
		if (value < 0) value = 0;
		if (value > (max = GetCtlMax(scrollCtl))) value = max;

		if (value != oldValue) {
			SetCtlValue(scrollCtl, value);
			teOffset = (*te)->viewRect.top - (*te)->destRect.top;
			if (value -= teOffset) TEScroll(0, -value, te);
		}
	}
}



/*****************************************************************************/



#pragma segment Controls
pascal void	HActionProc(ControlHandle scrollCtl, short part)
{
	short		lineHeight, delta, value, teOffset;
	short		oldValue, max;
	TEHandle	te;
	
	if (part) {						/* If it was actually in the control. */

		te = gActiveTEHndl;
		lineHeight = (*te)->lineHeight;

		switch (part) {
			case inUpButton:
			case inDownButton:		/* One line. */
				delta = lineHeight;
				break;
			case inPageUp:			/* One page. */
			case inPageDown:
				delta = (*te)->viewRect.right - (*te)->viewRect.left;
				if (delta > lineHeight) delta -= lineHeight;
				break;
		}
		if ( (part == inUpButton) || (part == inPageUp) )
			delta = -delta;		/* Reverse direction for an upper. */

		value = (oldValue = GetCtlValue(scrollCtl)) + delta;
		if (value < 0) value = 0;
		if (value > (max = GetCtlMax(scrollCtl))) value = max;

		if (value != oldValue) {
			SetCtlValue(scrollCtl, value);
			teOffset = (*te)->viewRect.left - (*te)->destRect.left;
			if (value -= teOffset) TEScroll(-value, 0, te);
		}
	}
}



/*****************************************************************************/



/* This function is called after an edit to make sure that there is no extra
** white space at the bottom of the viewRect.  If there are blank lines at
** the bottom of the viewRect, and there is text scrolled off the top of the
** viewRect, then the TextEdit control is scrolled to fill this space, or as
** much of it as possible.
*/

#pragma segment Controls
void	AdjustTEBottom(TEHandle teHndl)
{
	Rect	destRect, viewRect;
	short	botDiff, topDiff;

	destRect = (*teHndl)->destRect;
	viewRect = (*teHndl)->viewRect;
	destRect.bottom = destRect.top + CTEDocHeight(teHndl);

	botDiff = viewRect.bottom - destRect.bottom;
	if (botDiff > 0) {
		topDiff = viewRect.top - destRect.top;
		if (botDiff > topDiff) botDiff = topDiff;
		if (botDiff) TEScroll(0, botDiff, teHndl);
	}

}



/*****************************************************************************/



/* Bring the scrollbar values up to date with the current document position
** and length.
*/

#pragma segment Controls
void	AdjustScrollValues(TEHandle teHndl)
{
	short			scrollNum;
	ControlHandle	scrollCtl;

	for (scrollNum = 0; scrollNum < 2; ++scrollNum) {
		scrollCtl = CTEScrollFromTE(teHndl, scrollNum);
		if (scrollCtl)
			AdjustOneScrollValue(teHndl, scrollCtl, scrollNum);
	}
}



/*****************************************************************************/



/* Bring one scrollbar value up to date with the current document position
** and length.
*/

#pragma segment Controls
void	AdjustOneScrollValue(TEHandle teHndl, ControlHandle ctl, Boolean vert)
{
	Boolean	front;
	short	textPix, viewPix;
	short	max, oldMax, value, oldValue;

	front = ((*ctl)->contrlOwner == FrontWindow());

	oldValue = GetCtlValue(ctl);
	oldMax   = GetCtlMax(ctl);

	if (vert) {
		textPix = CTEDocHeight(teHndl);
		viewPix = (*teHndl)->viewRect.bottom - (*teHndl)->viewRect.top;
	}
	else {
		textPix = (*teHndl)->destRect.right - (*teHndl)->destRect.left;
		viewPix = (*teHndl)->viewRect.right - (*teHndl)->viewRect.left;
	}
	max = textPix - viewPix;

	if (max < 0) max = 0;
	if (max != oldMax) {
		if (front) SetCtlMax(ctl, max);
		else	   (*ctl)->contrlMax = max;
	}

	if (vert)
		value = (*teHndl)->viewRect.top  - (*teHndl)->destRect.top;
	else
		value = (*teHndl)->viewRect.left - (*teHndl)->destRect.left;

	if (value < 0)   value = 0;
	if (value > max) value = max;
	if (value != oldValue) {
		if (front) SetCtlValue(ctl, value);
		else	   (*ctl)->contrlValue = value;
	}
}



/*****************************************************************************/



#pragma segment Controls
Boolean	IsScrollBar(ControlHandle ctl)
{
	Rect			dummy;
	ControlHandle	dummyCtl;

	if (!ctl) return(false);

	if (!scrollProc) {
		SetRect(&dummy, 0, 0, 0, 0);
		dummyCtl = NewControl((*ctl)->contrlOwner, &dummy, "", 1, false, 0, 1, scrollBarProc, 0L);
		if (dummyCtl) {
			scrollProc = (*dummyCtl)->contrlDefProc;
			DisposeControl(dummyCtl);
		}
	}

	return((*ctl)->contrlDefProc == scrollProc);
}



